#
# WiKicker
#
#  ArticleDB.pm
#
#
# Copyright (C) 2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: ArticleDB.pm,v 1.1 2004/05/30 04:18:30 naney Exp $
#
#

package WiKicker::Diki::ArticleDB;

use 5.005;
use strict;
use Carp;
use DB_File::Lock;
use Fcntl qw(:flock O_RDWR O_RDONLY O_CREAT);
use WiKicker::Diki::DiaryArticle;
use WiKicker::Diki::DiaryDayList;
use WiKicker::Diki::KeywordList;
use WiKicker::Time;


#DB key:
#  d:YYYY-MM-DD-ID ... -> serialized DiaryArticle
#  k:ID            ... -> serialized KeywordArticle

sub new {
  my ($class, %param) = @_;

  unless (exists $param{file_name}) {
    croak "No file_name";
  }

  unless (exists $param{mode}) {
    croak "No mode";
  }

  unless ($param{mode} eq 'read' || $param{mode} eq 'write') {
    croak "Unknown mode: $param{mode}";
  }

  my $self = bless { file_name => $param{file_name},
                     mode => $param{mode},
                   }, $class;
  $self->_initialize;

  return $self;
}

sub DESTROY {
  my $self = shift;

  $self->_close_db;
}

sub _initialize {
  my $self = shift;

  $self->_open_db;
}

sub _open_db {
  my $self = shift;

  my $open_flag = O_RDWR|O_CREAT;
  #if ($self->{mode} eq 'read') {
  #  $open_flag = O_RDONLY|O_CREAT;
  #}

  my $locking = {
                 mode => $self->{mode},
                 nonblocking => $self->{mode} };
  my %db_hash;
  my $db_obj;
  for (1..120) {
    $db_obj = tie %db_hash, 'DB_File::Lock',
      $self->{file_name}, $open_flag, 0666, $DB_BTREE,
        $locking;
    if ($db_obj) {
      $self->{db_obj} = $db_obj;
      $self->{db_hash} = \%db_hash;
      return;
    }
    sleep(1);
  }
  croak "Can't open database: " . $self->{file_name};
}

sub _close_db {
  my $self = shift;

  undef $self->{db_obj};
  untie %{$self->{db_hash}};
}

sub _db_hash { shift->{db_hash} }
sub _db_obj  { shift->{db_obj} }

sub get_recent_diaries {
  my ($self, $size) = @_;

  my $key = 'd:end';
  my $value;
  my $status = $self->_db_obj->seq($key, $value, R_CURSOR);
  if ($status) {
    $self->_db_hash->{'d:end'} = 1;
    return WiKicker::Diki::DiaryDayList->new;
  }
  if ($key ne 'd:end') {
    $self->_db_hash->{'d:end'} = 1;
    return WiKicker::Diki::DiaryDayList->new;
  }
  my @result = ();

  my %date_map;
  for ($status = $self->_db_obj->seq($key, $value, R_PREV);
       $status == 0 && substr($key, 0, 2) eq 'd:';
       $status = $self->_db_obj->seq($key, $value, R_PREV)) {
    $key =~ /^d:(\d\d\d\d-\d\d-\d\d)/;
    next unless $1;
    $date_map{$1}++;
    last if (keys %date_map > $size);
    my $article = WiKicker::Diki::DiaryArticle->new(serialized => $value);
    push(@result, $article);
  }

  return WiKicker::Diki::DiaryDayList->new(article_list => \@result);
}

sub get_ten_day_diaries {
  my ($self, $range) = @_;

  $range =~ /^(\d\d\d\d)-(\d\d)-(.+)/;
  my ($y, $m, $ten) = ($1, $2, $3);
  my $start_day = '01';
  if ($ten eq 'b') {
    $start_day = '11';
  }
  elsif ($ten eq 'c') {
    $start_day = '21';
  }
  my $key = "d:$y-$m-$start_day";
  my $value;
  my @result;
  for (my $status = $self->_db_obj->seq($key, $value, R_CURSOR);
       $status == 0;
       $status = $self->_db_obj->seq($key, $value, R_NEXT)) {
    unless ($key =~ /^d:(\d\d\d\d)-(\d\d)-(\d\d)/) {
      last;
    }
    last if $y ne $1;
    last if $m ne $2;
    my $d = $3;
    last unless WiKicker::Time::day_in_ten_day($d, $ten);

    eval {
      my $article = WiKicker::Diki::DiaryArticle->new(serialized => $value);
      push(@result, $article);
    };
    if ($@) {
      warn $@;
    }
  }

  return WiKicker::Diki::DiaryDayList->new(article_list => \@result);
}

sub get_diaries {
  my ($self, $article_id_list) = @_;

  my @result;
  for my $article_id (@$article_id_list) {
    if (exists $self->_db_hash->{"d:$article_id"}) {
      my $article = WiKicker::Diki::DiaryArticle->new(serialized =>
                                                      $self->_db_hash->{"d:$article_id"}
                                                     );
      push(@result, $article);
    }
  }

  return WiKicker::Diki::DiaryDayList->new(article_list => \@result);
}

sub get_keywords {
  my ($self, $article_id_list) = @_;

  my @result;
  for my $article_id (@$article_id_list) {
    if (exists $self->_db_hash->{"k:$article_id"}) {
      my $article = WiKicker::Diki::KeywordArticle->new(serialized =>
                                                      $self->_db_hash->{"k:$article_id"}
                                                     );
      push(@result, $article);
    }
  }

  return WiKicker::Diki::KeywordList->new(article_list => \@result);
}


sub regist_diary_article {
  my ($self, $diary_article) = @_;

  $self->_db_hash->{'d:' . $diary_article->id} = $diary_article->serialize;
}

sub update_diary_article {
  my ($self, $diary_article) = @_;

  $self->_db_hash->{'d:' . $diary_article->id} = $diary_article->serialize;
}

sub regist_keyword_article {
  my ($self, $keyword_article) = @_;

  $self->_db_hash->{'k:' . $keyword_article->id} = $keyword_article->serialize;
}

sub update_keyword_article {
  my ($self, $keyword_article) = @_;

  $self->_db_hash->{'k:' . $keyword_article->id} = $keyword_article->serialize;
}


sub check_keyword_article {
  my ($self, $keyword_id, $modified) = @_;

  unless (exists $self->_db_hash->{'k:' . $keyword_id}) {
    return 'new';
  }

  my $value = $self->_db_hash->{'k:' . $keyword_id};
  my $article = WiKicker::Diki::KeywordArticle->new(serialized => $value);
  if ($modified > $article->last_modified) {
    return 'updated';
  }

  return 'not changed';
}

1;
