#
# WiKicker
#
#  TermDB.pm
#
#
# Copyright (C) 2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: TermDB.pm,v 1.1 2004/05/30 04:18:30 naney Exp $
#
#

package WiKicker::Diki::TermDB;

use 5.005;
use strict;
use Carp;
use DB_File;
use Storable qw(freeze thaw);
use Fcntl qw(O_RDWR O_CREAT);

#IFDEBUG: use Carp::Assert;

# t@size           | num
# t@mod            : time
# t:<term>         | TermRecord
# d:<diary id>     | DiaryRecord
# k:<keyword id>   | KeywordRecord
# T#<term index>   | string
# D#<diary index>  | string
# K#<keyword index>| string
# T@next           | num
# D@next           | num
# K@next           | num

# DiaryRecord
# [
#  diary index,
#  {explicit_term_index_1 => 1, explicit_term_index_2 => 1, ...}
#  {implicit_term_index_1 => 1, implicit_term_index_2 => 1, ...}
# ]

# KeywordRecord
# [
#  keyword index,
#  {explicit_term_index_1 => 1, explicit_term_index_2 => 1, ...}
#  {implicit_term_index_1 => 1, implicit_term_index_2 => 1, ...}
#  {defined_term_index_1  => 1, defined_term_index_2  => 1, ...}
# ]

# TermRecord
# [
#  term index,
#  {explicit_diary_index_1 => 1, explicit_diary_index_2 => 1, ...}
#  {implicit_diary_index_1 => 1, implicit_diary_index_2 => 1, ...}
#  {explicit_keyword_index_1 => 1, explicit_keyword_index_2 => 1, ...}
#  {implicit_keyword_index_1 => 1, implicit_keyword_index_2 => 1, ...}
#  {defined_keyword_index_1 => 1, defined_keyword_index_2 => 1, ...}
# ]

sub new {
  my ($class, %param) = @_;

  unless (exists $param{file_name}) {
    croak "No file_name";
  }

  my $self = bless { file_name => $param{file_name} }, $class;
  $self->_initialize;

  return $self;
}

sub DESTROY {
  my $self = shift;

  $self->_close_db;
}

sub _initialize {
  my $self = shift;

  $self->_open_db;
}

sub _open_db {
  my $self = shift;

  my %db_hash;
  my $db_obj = tie %db_hash, 'DB_File',
    $self->{file_name}, O_RDWR|O_CREAT, 0666, $DB_BTREE;
  if ($db_obj) {
    $self->{db_obj} = $db_obj;
    $self->{db_hash} = \%db_hash;
    return;
  }
  else {
    croak "Can't open database: " . $self->{file_name} . ": $!";
  }
}

sub hash { shift->{db_hash} }

sub _close_db {
  my $self = shift;

  undef $self->{db_obj};
  untie %{$self->{db_hash}};
}


sub term_list {
  my $self = shift;

  if (exists $self->hash->{'t@size'}) {
    return $self->_get_term_list_with_size($self->hash->{'t@size'});
  }

  my @term_list = ();
  my $size = $self->_get_term_list_from_hash(\@term_list);
  $self->hash->{'t@size'} = $size;
  $self->{term_list} = \@term_list;

  return ($self->{term_list},
          $self->_get_term_set_modified);
}

sub _get_term_list_with_size {
  my ($self, $size) = @_;

  if ($size == 0) {
    $self->{term_list} = [];
    return ($self->{term_list}, $self->_get_term_set_modified);
  }

  my @term_list = ();
  $#term_list = $size;
  my $new_size = $self->_get_term_list_from_hash(\@term_list);
  $self->{term_list} = \@term_list;
  if ($new_size != $size) {
    $self->hash->{'t@size'} = $new_size;
    $#term_list = $new_size;
  }
  return ($self->{term_list}, $self->_get_term_set_modified);
}

sub _get_term_list_from_hash {
  my ($self, $term_list_ref) = @_;

  my $i = 0;
  # @TODO don't foreach all key
  foreach (keys %{$self->hash}) {
    my $key = $_;
    if (substr($key, 0, 2) eq 't:') {
      $term_list_ref->[$i] = substr($key, 2);
      $i++;
    }
  }
  return $i;
}

sub _exists_term {
  my ($self, $term) = @_;

  return exists $self->hash->{"t:$term"};
}

sub _term_index_to_term {
  my ($self, $term_index) = @_;

  #IFDEBUG: affirm {
  #IFDEBUG:   warn "unknown term_index $term_index"
  #IFDEBUG:     unless exists $self->hash->{'T#' . $term_index};
  #IFDEBUG:   exists $self->hash->{'T#' . $term_index};
  #IFDEBUG: };
  return $self->hash->{'T#' . $term_index};
}

sub _get_term_record {
  my ($self, $term) = @_;

  #IFDEBUG: affirm { exists $self->hash->{"t:$term"} };
  #IFDEBUG: affirm { defined $self->hash->{"t:$term"} };
  return WiKicker::Diki::TermDB::TermRecord->new(serialized =>
       $self->hash->{"t:$term"});
}

sub _set_term_record {
  my ($self, $term, $record) = @_;

  #IFDEBUG: affirm { my $i = $record->index;
  #IFDEBUG:          exists $self->hash->{'T#' . $i} &&
  #IFDEBUG:            $self->hash->{'T#' . $i} eq $term;
  #IFDEBUG: };
  $self->hash->{"t:$term"} = $record->serialize;
}

sub _new_term_record {
  my ($self, $term) = @_;

  my $new_index = $self->_get_new_term_index($term);
  return WiKicker::Diki::TermDB::TermRecord->new(index => $new_index);
}

sub _new_diary_record {
  my ($self, $diary_id) = @_;

  my $index = $self->_get_new_diary_index($diary_id);
  return WiKicker::Diki::TermDB::DiaryRecord->new(index => $index);
}

sub _new_keyword_record {
  my ($self, $keyword_id) = @_;

  my $index = $self->_get_new_keyword_index($keyword_id);
  return WiKicker::Diki::TermDB::KeywordRecord->new(index => $index);
}


sub _get_diary_record {
  my ($self, $diary_id) = @_;
  return WiKicker::Diki::TermDB::DiaryRecord->new(serialized => $self->hash->{'d:' . $diary_id});
}

sub _set_diary_record {
  my ($self, $diary_id, $record) = @_;

  $self->hash->{'d:' . $diary_id} = $record->serialize;
}

sub _set_keyword_record {
  my ($self, $keyword_id, $record) = @_;

  $self->hash->{'k:' . $keyword_id} = $record->serialize;
}

sub _get_keyword_record {
  my ($self, $keyword_id) = @_;
  return WiKicker::Diki::TermDB::KeywordRecord->new(serialized => $self->hash->{'k:' . $keyword_id});
}

sub _get_diary_id_from_index {
  my ($self, $index) = @_;

  #IFDEBUG: affirm { exists $self->hash->{'D#' . $index} };
  return $self->hash->{'D#' . $index};
}

sub _get_keyword_id_from_index {
  my ($self, $index) = @_;

  #IFDEBUG: affirm { exists $self->hash->{'K#' . $index} };
  return $self->hash->{'K#' . $index};
}

sub get_recent_diaries {
  my ($self, $name, $size) = @_;

  if ($self->_exists_term($name)) {
    my $record = $self->_get_term_record($name);
    my $refered_diary_index_list = $record->refered_diary_index_list;
    my @result = ();
    for my $index (@$refered_diary_index_list) {
      push(@result, $self->_get_diary_id_from_index($index));
    }
    return \@result;
  }

  return [];
}

sub get_keyword_articles {
  my ($self, $term) = @_;

  if ($self->_exists_term($term)) {
    my $record = $self->_get_term_record($term);
    my $keyword_index_list = $record->keyword_index_list;
    my @result = ();
    for my $index (@$keyword_index_list) {
      push(@result, $self->_get_keyword_id_from_index($index));
    }
    return \@result;
  }

  return [];
}

sub regist_diary_article {
  my ($self, $diary_article) = @_;

  my $diary_record = $self->_new_diary_record($diary_article->id);

  my $term_size_delta = 0;
  for my $term (@{$diary_article->explicit_term_list}) {
    my ($term_index, $new) = $self->_set_diary_term($term,
                                                    $diary_record->index,
                                                    'explicit');
    $diary_record->set_explicit_term($term_index);
    $term_size_delta += $new;
  }

  for my $term (@{$diary_article->implicit_term_list}) {
    my ($term_index, $new) = $self->_set_diary_term($term,
                                                    $diary_record->index,
                                                    'implicit');
    $diary_record->set_implicit_term($term_index);
    $term_size_delta += $new;
  }

  if ($term_size_delta != 0) {
    $self->hash->{'t@size'} += $term_size_delta;
    $self->_set_term_set_modified(time);
  }

  $self->_set_diary_record($diary_article->id, $diary_record);

  return;
}

sub regist_keyword_article {
  my ($self, $keyword_article) = @_;

  my $keyword_record = $self->_new_keyword_record($keyword_article->id);

  my $term_size_delta = 0;
  for my $term (@{$keyword_article->explicit_term_list}) {
    my ($term_index, $new) = $self->_set_keyword_term($term,
                                                      $keyword_record->index,
                                                      'explicit');
    $keyword_record->set_explicit_term($term_index);
    $term_size_delta += $new;
  }

  for my $term (@{$keyword_article->implicit_term_list}) {
    my ($term_index, $new) = $self->_set_keyword_term($term,
                                                      $keyword_record->index,
                                                      'implicit');
    $keyword_record->set_implicit_term($term_index);
    $term_size_delta += $new;
  }

  for my $term (@{$keyword_article->defined_term_list}) {
    my ($term_index, $new) = $self->_set_keyword_term($term,
                                                      $keyword_record->index,
                                                      'define');
    $keyword_record->set_defined_term($term_index);
    $term_size_delta += $new;
  }

  if ($term_size_delta != 0) {
    $self->hash->{'t@size'} += $term_size_delta;
    $self->_set_term_set_modified(time);
  }

  $self->_set_keyword_record($keyword_article->id, $keyword_record);

  return;
}


sub update_diary_article_entry {
  my ($self, $diary_article) = @_;

  my $new_explicit_term_list = $diary_article->explicit_term_list;
  my $new_implicit_term_list = $diary_article->implicit_term_list;
  my $diary_record = $self->_get_diary_record($diary_article->id);
  my $old_explicit_term_set =
    $self->make_term_to_term_index_map($diary_record->get_explicit_term_index_list);
  my $old_implicit_term_set =
    $self->make_term_to_term_index_map($diary_record->get_implicit_term_index_list);

  # added
  my $term_set_changed = 0;
  my $term_size_delta = 0;
  for my $new_term (@$new_explicit_term_list) {
    my ($term_index, $new) = $self->_set_diary_term($new_term,
                                                    $diary_record->index,
                                                    'explicit'
                                                   );
    $diary_record->set_explicit_term($term_index);
    $term_size_delta += $new;
    $term_set_changed |= $new;
    delete $old_explicit_term_set->{$new_term};
  }
  for my $new_term (@$new_implicit_term_list) {
    my ($term_index, $new) = $self->_set_diary_term($new_term,
                                                    $diary_record->index,
                                                   'implicit');
    $diary_record->set_implicit_term($term_index);
    $term_size_delta += $new;
    $term_set_changed |= $new;
    delete $old_implicit_term_set->{$new_term};
  }

  # removed?
  for my $old_term (keys %$old_explicit_term_set) {

    $diary_record->remove_explicit_term($old_explicit_term_set->{$old_term});
    my $deleted = $self->_remove_explicit_term($old_term,
                                               $diary_record->index);
    if ($deleted) {
      $term_size_delta -= 1;
      $term_set_changed = 1;
    }
  }
  for my $old_term (keys %$old_implicit_term_set) {

    $diary_record->remove_implicit_term($old_implicit_term_set->{$old_term});
    my $deleted = $self->_remove_implicit_term($old_term,
                                               $diary_record->index);
    if ($deleted) {
      $term_size_delta -= 1;
      $term_set_changed = 1;
    }
  }

  if ($term_set_changed) {
    $self->hash->{'t@size'} += $term_size_delta;
    $self->_set_term_set_modified(time);
  }
  $self->_set_diary_record($diary_article->id, $diary_record);
}

sub update_keyword_article {
  my ($self, $keyword_article) = @_;

  my $new_explicit_term_list = $keyword_article->explicit_term_list;
  my $new_implicit_term_list = $keyword_article->implicit_term_list;
  my $new_defined_term_list = $keyword_article->defined_term_list;
  my $keyword_record = $self->_get_keyword_record($keyword_article->id);
  my $old_explicit_term_set =
    $self->make_term_to_term_index_map($keyword_record->get_explicit_term_index_list);
  my $old_implicit_term_set =
    $self->make_term_to_term_index_map($keyword_record->get_implicit_term_index_list);
  my $old_defined_term_set =
    $self->make_term_to_term_index_map($keyword_record->get_defined_term_index_list);

  # added
  my $term_set_changed = 0;
  my $term_size_delta = 0;
  for my $new_term (@$new_explicit_term_list) {
    my ($term_index, $new) = $self->_set_keyword_term($new_term,
                                                    $keyword_record->index,
                                                    'explicit'
                                                   );
    $keyword_record->set_explicit_term($term_index);
    $term_size_delta += $new;
    $term_set_changed |= $new;
    delete $old_explicit_term_set->{$new_term};
  }
  for my $new_term (@$new_implicit_term_list) {
    my ($term_index, $new) = $self->_set_keyword_term($new_term,
                                                    $keyword_record->index,
                                                   'implicit');
    $keyword_record->set_implicit_term($term_index);
    $term_size_delta += $new;
    $term_set_changed |= $new;
    delete $old_implicit_term_set->{$new_term};
  }

  for my $new_term (@$new_defined_term_list) {
    my ($term_index, $new) = $self->_set_keyword_term($new_term,
                                                    $keyword_record->index,
                                                   'define');
    $keyword_record->set_defined_term($term_index);
    $term_size_delta += $new;
    $term_set_changed |= $new;
    delete $old_defined_term_set->{$new_term};
  }

  # removed?
  for my $old_term (keys %$old_explicit_term_set) {

    $keyword_record->remove_explicit_term($old_explicit_term_set->{$old_term});
    my $deleted = $self->_remove_explicit_term($old_term,
                                               $keyword_record->index);
    if ($deleted) {
      $term_size_delta -= 1;
      $term_set_changed = 1;
    }
  }
  for my $old_term (keys %$old_implicit_term_set) {

    $keyword_record->remove_implicit_term($old_implicit_term_set->{$old_term});
    my $deleted = $self->_remove_implicit_term($old_term,
                                               $keyword_record->index);
    if ($deleted) {
      $term_size_delta -= 1;
      $term_set_changed = 1;
    }
  }

  for my $old_term (keys %$old_defined_term_set) {

    $keyword_record->remove_defined_term($old_defined_term_set->{$old_term});
    my $deleted = $self->_remove_defined_term($old_term,
                                              $keyword_record->index);
    if ($deleted) {
      $term_size_delta -= 1;
      $term_set_changed = 1;
    }
  }


  if ($term_set_changed) {
    $self->hash->{'t@size'} += $term_size_delta;
    $self->_set_term_set_modified(time);
  }
  $self->_set_keyword_record($keyword_article->id, $keyword_record);
}


sub make_term_to_term_index_map {
  my ($self, $term_index_list) = @_;

  my %result;
  for my $term_index (@$term_index_list) {
    #IFDEBUG: affirm { defined $term_index && $term_index };
    my $term = $self->_term_index_to_term($term_index);
    $result{$term} = $self->_get_term_record($term)->index;
  }
  return \%result;
}

sub _set_diary_term {
  my ($self, $term, $diary_index, $type) = @_;

  #IFDEF: affirm { $term ne '' }
  my $new = 0;
  my $term_record;
  if ($self->_exists_term($term)) {
    $term_record = $self->_get_term_record($term);
  }
  else {
    $new = 1;
    $term_record = $self->_new_term_record($term);
  }
  $term_record->set_diary($type, $diary_index);
  $self->_set_term_record($term, $term_record);

  return ($term_record->index, $new);
}

sub _set_keyword_term {
  my ($self, $term, $keyword_index, $type) = @_;

  #IFDEF: affirm { $term ne '' }
  my $new = 0;
  my $term_record;
  if ($self->_exists_term($term)) {
    $term_record = $self->_get_term_record($term);
  }
  else {
    $new = 1;
    $term_record = $self->_new_term_record($term);
  }
  $term_record->set_keyword($type, $keyword_index);
  $self->_set_term_record($term, $term_record);

  return ($term_record->index, $new);
}

sub _remove_defined_term {
  my ($self, $term, $keyword_index) = @_;

  #IFDEF: affirm { $term ne '' }
  if ($self->_exists_term($term)) {
    my $term_record = $self->_get_term_record($term);
    $term_record->remove_keyword($keyword_index, 'define');
  }
}

sub _get_new_diary_index {
  my ($self, $id) = @_;

  return $self->_get_new_index('D', $id);
}

sub _get_new_keyword_index {
  my ($self, $id) = @_;

  return $self->_get_new_index('K', $id);
}

sub _get_new_term_index {
  my ($self, $term) = @_;

  return $self->_get_new_index('T', $term);
}

sub _get_new_index {
  my ($self, $type, $id) = @_;

  my $index = 1;
  if (exists $self->hash->{$type . '@next'}) {
    $index = $self->hash->{$type . '@next'};
  }
  $self->hash->{$type . '@next'} = $index + 1;
  $self->hash->{$type . '#' . $index} = $id;

  return $index;
}


sub _get_term_set_modified {
  my $self = shift;

  if (exists $self->{'t@mod'}) {
    return $self->{'t@mod'};
  }

  if (exists $self->hash->{'t@mod'}) {
    return $self->{'t@mod'}
      = $self->hash->{'t@mod'};
  }

  $self->_set_term_set_modified(time);

  return $self->{'t@mod'};
}

sub _set_term_set_modified {
  my ($self, $time) = @_;

  $self->hash->{'t@mod'} = $self->{'t@mod'} = $time;
}

package WiKicker::Diki::TermDB::DiaryRecord;
use strict;
use Storable qw(freeze thaw);
use Carp;

sub new {
  my ($class, %param) = @_;

  if (exists $param{serialized}) {
    my $self = thaw($param{serialized});
    bless $self, $class;
    return $self;
  }

  else {
    croak "no index" unless $param{index};
  }

  my $self = bless [ $param{index}, {}, {} ], $class;

  return $self;
}

sub serialize {
  my $self = shift;

  return freeze($self);
}

sub index { shift->[0] }

sub set_explicit_term {
  my ($self, $term_index) = @_;

  $self->[1]->{$term_index} = 1;
}

sub set_implicit_term {
  my ($self, $term_index) = @_;

  $self->[2]->{$term_index} = 1;
}

sub get_explicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[1]};
  return \@result;
}

sub get_implicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[2]};
  return \@result;
}

package WiKicker::Diki::TermDB::KeywordRecord;
use strict;
use Storable qw(freeze thaw);
use Carp;

sub new {
  my ($class, %param) = @_;

  if (exists $param{serialized}) {
    my $self = thaw($param{serialized});
    bless $self, $class;
    return $self;
  }
  else {
    croak "no index" unless $param{index};
  }

  my $self = bless [ $param{index}, {}, {} ,{} ], $class;

  return $self;
}

sub serialize {
  my $self = shift;

  return freeze($self);
}


sub index { shift->[0] }

sub set_explicit_term {
  my ($self, $term_index) = @_;

  $self->[1]->{$term_index} = 1;
}

sub set_implicit_term {
  my ($self, $term_index) = @_;

  $self->[2]->{$term_index} = 1;
}

sub set_defined_term {
  my ($self, $term_index) = @_;

  $self->[3]->{$term_index} = 1;
}

sub get_explicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[1]};
  return \@result;
}

sub get_implicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[2]};
  return \@result;
}

sub get_defined_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[3]};
  return \@result;
}

package WiKicker::Diki::TermDB::TermRecord;
use strict;
use Storable qw(freeze thaw);
use Carp;

sub new {
  my ($class, %param) = @_;

  if (exists $param{serialized}) {
    my $self = thaw($param{serialized});
    bless $self, $class;
    return $self;
  }
  else {
    croak "no index" unless $param{index};
  }

  my $self = bless [ $param{index}, {}, {}, {}, {}, {} ], $class;

  return $self;
}

sub serialize {
  my $self = shift;

  return freeze($self);
}

sub index { shift->[0] }

sub set_diary {
  my ($self, $type, $diary_index) = @_;

  if ($type eq 'explicit') {
    $self->[1]->{$diary_index} = 1;
  }
  elsif ($type eq 'implicit') {
    $self->[2]->{$diary_index} = 1;
  }
  else {
    die;
  }
}

sub set_keyword {
  my ($self, $type, $keyword_index) = @_;

  if ($type eq 'explicit') {
    $self->[3]->{$keyword_index} = 1;
  }
  elsif ($type eq 'implicit') {
    $self->[4]->{$keyword_index} = 1;
  }
  elsif ($type eq 'define') {
    $self->[5]->{$keyword_index} = 1;
  }
  else {
    die;
  }
}

sub remove_keyword {
  my ($self, $keyword_index, $type) = @_;

  if ($type eq 'explicit') {
    delete $self->[3]->{$keyword_index};
  }
  elsif ($type eq 'implicit') {
    delete $self->[4]->{$keyword_index};
  }
  elsif ($type eq 'define') {
    delete $self->[5]->{$keyword_index};
  }
  else {
    die;
  }
}

sub refered_diary_index_list {
  my $self = shift;

  my %result = ();
  for my $index (keys %{$self->[1]}) {
    $result{$index}++;
  }
  for my $index (keys %{$self->[2]}) {
    $result{$index}++;
  }
  return [ keys %result ];
}

sub keyword_index_list {
  my $self = shift;

  my @result = keys %{$self->[5]};
  return \@result;
}



1;
